#!/usr/bin/env python
#
#    Jools -- a graphical puzzle game in the Tetris tradition
#    
#    Copyright (C) 2002-2003 Paul Pelzl
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

# __init__.py
#
# Primary game control loop.
#

import os, random, sys

try:
   import pygame
except ImportError:
   sys.exit("Jools requires Pygame, which does not appear to be installed.")


from pygame.locals import *

from globals import *
from joolsutils import *

from initialize import *
import intro, grid, highscore, inputbox


def gameExit():
   pygame.quit()
   sys.exit("Exiting Jools...")



pointsTextBackground = joolsFont.render("Points: 100000000",
   1, (1, 1, 1), (1, 1, 1))
movesTextBackground  = joolsFont.render("Moves Available: 999",
   1, (1, 1, 1), (1, 1, 1))
levelTextBackground = joolsFont.render("Level: 10000",
   1, (1, 1, 1), (1, 1, 1))
numJoolsTextBackground = joolsFont.render("Jools Cleared: 10000000",
   1, (1, 1, 1), (1, 1, 1))

joolsClock = pygame.time.Clock()


# Handle initial player configuration, game choice, etc.
while 1:

   # Create an empty grid.
   joolsGrid = grid.Grid()

   timeTrial = intro.showIntro()

   gameOver = 0
   level = 1
   timerCorrection = 2.0
   joolsGrid.totalPoints = 0
   joolsGrid.joolsThisLevel = 0
   joolsGrid.totalJools  = 0
   
   # Play a complete game with levels, etc.
   while not gameOver:
      # Start the game (let the jools fall)
      joolsGrid.new(timeTrial)  # Fill empty grid with jools
      while joolsGrid.updateAll(UPDATEFALLING):
         pygame.event.pump() 
         if pygame.event.peek(QUIT):
            gameExit()
         pygame.event.get()  # Clear the event queue
         # Throttle the framerate a bit as the jools fall
         joolsClock.tick(FALLING_TARGET_FPS)
         joolsGrid.cleanUpdated()
         joolsGrid.renderUpdated()
         pygame.display.flip()
      getNewGrid = 0

      screen.blit(levelTextBackground, (5, 220))
      levelText = joolsFont.render("Level: " + str(level),
         1, (250, 0, 0), (1, 1, 1) )
      screen.blit(levelText, (5, 220))

      screen.blit(pointsTextBackground, (5, 280))
      pointsText = joolsFont.render("Points: " + str(joolsGrid.totalPoints),
         1, (250, 0, 0), (1, 1, 1) )
      screen.blit(pointsText, (5, 280))

      screen.blit(numJoolsTextBackground, (5, 340))
      numJoolsText = joolsFont.render("Jools Cleared: " + str(joolsGrid.totalJools),
         1, (250, 0, 0), (1, 1, 1) )
      screen.blit(numJoolsText, (5, 340))

      # Main game loop
      gridIsOld = 0     # Points should be tallied only "if the grid is old",
                        # after the fresh grid has settled and initial matches have
                        # been removed.
      while not getNewGrid and not gameOver:
         # Remove jools and let them fall
         matches = joolsGrid.getMatches(gridIsOld)
         while len(matches) > 0:
            screen.blit(pointsTextBackground, (5, 280))
            pointsText = joolsFont.render("Points: " + str(joolsGrid.totalPoints),
               1, (250, 0, 0), (1, 1, 1) )
            screen.blit(pointsText, (5, 280))
            if timeTrial:
               joolsGrid.timer.boost(timerCorrection*float(joolsGrid.pointsToAdd))
            joolsGrid.removeJools(matches, gridIsOld)

            screen.blit(numJoolsTextBackground, (5, 340))
            numJoolsText = joolsFont.render("Jools Cleared: " + str(joolsGrid.totalJools),
               1, (250, 0, 0), (1, 1, 1) )
            screen.blit(numJoolsText, (5, 340))

            while joolsGrid.updateAll(UPDATEFALLING) and not gameOver:
               pygame.event.pump()
               if pygame.event.peek(QUIT):
                  gameExit()
               pygame.event.get()  # Clear the event queue
               # Throttle the framerate a bit as the jools fall
               joolsClock.tick(FALLING_TARGET_FPS)
               joolsGrid.cleanUpdated()
               joolsGrid.renderUpdated()
               pygame.display.flip()
            matches = joolsGrid.getMatches(gridIsOld)

         gridIsOld = 1

         if joolsGrid.joolsThisLevel > JOOLS_PER_LEVEL:
            # New level!
            joolsGrid.joolsThisLevel = 0
            level += 1
            if timeTrial:
               joolsGrid.timer.timer_dec += TIMER_DECREMENT_MOD

            screen.blit(levelTextBackground, (5, 220))
            levelText = joolsFont.render("Level: " + str(level),
               1, (250, 0, 0), (1, 1, 1) )
            screen.blit(levelText, (5, 220))

            getNewGrid = 1
            screen.blit(levelUpImage, (SCREENW-(8*JOOLSIZE) + ((8*JOOLSIZE)-levelUpRect.width)/2, 
               (SCREENH-(8*JOOLSIZE) + ((8*JOOLSIZE)-levelUpRect.height)/2)))
            pygame.display.update()
            pygame.time.wait(1500)
            screen.blit(background, (SCREENW-(8*JOOLSIZE), SCREENH-(8*JOOLSIZE)))

         # Process input (Most of the cycles go here)
         if not getNewGrid:
            joolsGrid.updateMoves()
            screen.blit(movesTextBackground, (5, 400))
            movesText = joolsFont.render("Moves Available: " + 
               str(len(joolsGrid.availableMoves)), 1, (250, 0, 0), (1, 1, 1) )
            screen.blit(movesText, (5, 400))
            pygame.display.flip()
            if len(joolsGrid.availableMoves) > 0:
               while not (joolsGrid.hasSwaps) and not gameOver:
                  pygame.event.pump()
                  if pygame.event.peek([QUIT]):
                     gameExit()
                  if joolsGrid.processInput():
                     gameOver = 1
                  pygame.event.get()  # clear the event queue
                  # Throttle the framerate a bit on rotations
                  joolsClock.tick(ROTATING_TARGET_FPS)
                  if timeTrial:
                     joolsGrid.timer.tick()
                  if timeTrial and joolsGrid.timer.time <= 0.0:
                     gameOver = 1
                  if joolsGrid.updateAll(UPDATEROTATION):
                     #pygame.time.wait(ROTATIONDELAY)
                     joolsGrid.cleanUpdated()
                     joolsGrid.renderUpdated()

               runTimer = 0
               if joolsGrid.hasFalseSwaps:
                  runTimer = 1
               # Handle jool swapping
               while joolsGrid.updateAll(UPDATESWAPS) and not gameOver:
                  pygame.event.pump()
                  if pygame.event.peek(QUIT):
                     gameExit()
                  pygame.event.get()
                  # Throttle the framerate a bit on jool swapping
                  joolsClock.tick(SWAPPING_TARGET_FPS)
                  # If the jools are being swapped illegally,
                  # then the clock still runs.
                  if timeTrial and runTimer:
                     joolsGrid.timer.tick()
                  joolsGrid.cleanUpdated()
                  joolsGrid.renderUpdated()
                  pygame.display.flip()
            else:
               # No moves remaining!
               if timeTrial:
                  getNewGrid = 1
                  screen.blit(noMovesImage, (SCREENW-(8*JOOLSIZE) + ((8*JOOLSIZE)-noMovesRect.width)/2, 
                     (SCREENH-(8*JOOLSIZE) + ((8*JOOLSIZE)-noMovesRect.height)/2)))
                  pygame.display.update()
                  pygame.time.wait(1500)
                  screen.blit(background, (640-(8*JOOLSIZE), 480-(8*JOOLSIZE)))
               else:
                  gameOver = 1

   
   screen.blit(gameOverImage, (SCREENW-(8*JOOLSIZE) + ((8*JOOLSIZE)-gameOverRect.width)/2, 
      (SCREENH-(8*JOOLSIZE) + ((8*JOOLSIZE)-gameOverRect.height)/2)))
   pygame.display.update()
   # Add a hint of suspense
   pygame.time.wait(1000)

   # Is this a high score?
   if highscore.isHighScore(joolsGrid.totalPoints, timeTrial):
      HSImageLocation = (SCREENW-(8*JOOLSIZE) + ((8*JOOLSIZE)-highScoreRect.width)/2,
         SCREENH-(8*JOOLSIZE) + ((8*JOOLSIZE)-gameOverRect.height)/2 + gameOverRect.height - 30)
      screen.blit(highScoreImage, HSImageLocation)
      pygame.display.update()
      
      boxLocation = (HSImageLocation[0] + highScoreRect.width/2, HSImageLocation[1] + highScoreRect.height + 15)
      name = inputbox.ask(screen, (200,200,200), (1,1,1), " Name", boxLocation)

      rank = highscore.addScore(name, level, joolsGrid.totalPoints, timeTrial)
      highscore.showScores(rank, timeTrial)

   else:
      pygame.time.wait(5000)
      

